# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/experimental/workflow/protocols/director.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5openfl/experimental/workflow/protocols/director.proto\x12%openfl.experimental.workflow.director\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\"+\n\x15SendConnectionRequest\x12\x12\n\nenvoy_name\x18\x01 \x01(\t\"#\n\x0fRequestAccepted\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"2\n\x15WaitExperimentRequest\x12\x19\n\x11\x63ollaborator_name\x18\x01 \x01(\t\"1\n\x16WaitExperimentResponse\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"N\n\x18GetExperimentDataRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\x12\x19\n\x11\x63ollaborator_name\x18\x02 \x01(\t\"0\n\x0e\x45xperimentData\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x10\n\x08\x65xp_data\x18\x02 \x01(\x0c\"G\n\x18UpdateEnvoyStatusRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x15is_experiment_running\x18\x02 \x01(\x08\"S\n\x19UpdateEnvoyStatusResponse\x12\x36\n\x13health_check_period\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x8a\x01\n\x0e\x45xperimentInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x12\x63ollaborator_names\x18\x02 \x03(\t\x12N\n\x0f\x65xperiment_data\x18\x03 \x01(\x0b\x32\x35.openfl.experimental.workflow.director.ExperimentData\"*\n\x18SetNewExperimentResponse\x12\x0e\n\x06status\x18\x01 \x01(\x08\"\xcf\x01\n\tEnvoyInfo\x12\x12\n\nenvoy_name\x18\x01 \x01(\t\x12\x17\n\x0f\x65xperiment_name\x18\x02 \x01(\t\x12\x11\n\tis_online\x18\x03 \x01(\x08\x12\x1d\n\x15is_experiment_running\x18\x04 \x01(\x08\x12\x30\n\x0clast_updated\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\x0evalid_duration\x18\x06 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x12\n\x10GetEnvoysRequest\"Z\n\x11GetEnvoysResponse\x12\x45\n\x0b\x65nvoy_infos\x18\x01 \x03(\x0b\x32\x30.openfl.experimental.workflow.director.EnvoyInfo\"\x15\n\x13GetFlowStateRequest\"=\n\x14GetFlowStateResponse\x12\x11\n\tcompleted\x18\x01 \x01(\x08\x12\x12\n\nflspec_obj\x18\x02 \x01(\x0c\"\x14\n\x12SendRuntimeRequest\"*\n\x16RuntimeRequestResponse\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"5\n\x1aGetExperimentStdoutRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"]\n\x1bGetExperimentStdoutResponse\x12\x15\n\rstdout_origin\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x14\n\x0cstdout_value\x18\x03 \x01(\t2\xae\n\n\x08\x44irector\x12\x90\x01\n\x16\x45nvoyConnectionRequest\x12<.openfl.experimental.workflow.director.SendConnectionRequest\x1a\x36.openfl.experimental.workflow.director.RequestAccepted\"\x00\x12\x8f\x01\n\x0eWaitExperiment\x12<.openfl.experimental.workflow.director.WaitExperimentRequest\x1a=.openfl.experimental.workflow.director.WaitExperimentResponse\"\x00\x12\x8f\x01\n\x11GetExperimentData\x12?.openfl.experimental.workflow.director.GetExperimentDataRequest\x1a\x35.openfl.experimental.workflow.director.ExperimentData\"\x00\x30\x01\x12\x98\x01\n\x11UpdateEnvoyStatus\x12?.openfl.experimental.workflow.director.UpdateEnvoyStatusRequest\x1a@.openfl.experimental.workflow.director.UpdateEnvoyStatusResponse\"\x00\x12\x8e\x01\n\x10SetNewExperiment\x12\x35.openfl.experimental.workflow.director.ExperimentInfo\x1a?.openfl.experimental.workflow.director.SetNewExperimentResponse\"\x00(\x01\x12\x80\x01\n\tGetEnvoys\x12\x37.openfl.experimental.workflow.director.GetEnvoysRequest\x1a\x38.openfl.experimental.workflow.director.GetEnvoysResponse\"\x00\x12\x89\x01\n\x0cGetFlowState\x12:.openfl.experimental.workflow.director.GetFlowStateRequest\x1a;.openfl.experimental.workflow.director.GetFlowStateResponse\"\x00\x12\x8c\x01\n\x0e\x43onnectRuntime\x12\x39.openfl.experimental.workflow.director.SendRuntimeRequest\x1a=.openfl.experimental.workflow.director.RuntimeRequestResponse\"\x00\x12\xa0\x01\n\x13GetExperimentStdout\x12\x41.openfl.experimental.workflow.director.GetExperimentStdoutRequest\x1a\x42.openfl.experimental.workflow.director.GetExperimentStdoutResponse\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfl.experimental.workflow.protocols.director_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SENDCONNECTIONREQUEST']._serialized_start=161
  _globals['_SENDCONNECTIONREQUEST']._serialized_end=204
  _globals['_REQUESTACCEPTED']._serialized_start=206
  _globals['_REQUESTACCEPTED']._serialized_end=241
  _globals['_WAITEXPERIMENTREQUEST']._serialized_start=243
  _globals['_WAITEXPERIMENTREQUEST']._serialized_end=293
  _globals['_WAITEXPERIMENTRESPONSE']._serialized_start=295
  _globals['_WAITEXPERIMENTRESPONSE']._serialized_end=344
  _globals['_GETEXPERIMENTDATAREQUEST']._serialized_start=346
  _globals['_GETEXPERIMENTDATAREQUEST']._serialized_end=424
  _globals['_EXPERIMENTDATA']._serialized_start=426
  _globals['_EXPERIMENTDATA']._serialized_end=474
  _globals['_UPDATEENVOYSTATUSREQUEST']._serialized_start=476
  _globals['_UPDATEENVOYSTATUSREQUEST']._serialized_end=547
  _globals['_UPDATEENVOYSTATUSRESPONSE']._serialized_start=549
  _globals['_UPDATEENVOYSTATUSRESPONSE']._serialized_end=632
  _globals['_EXPERIMENTINFO']._serialized_start=635
  _globals['_EXPERIMENTINFO']._serialized_end=773
  _globals['_SETNEWEXPERIMENTRESPONSE']._serialized_start=775
  _globals['_SETNEWEXPERIMENTRESPONSE']._serialized_end=817
  _globals['_ENVOYINFO']._serialized_start=820
  _globals['_ENVOYINFO']._serialized_end=1027
  _globals['_GETENVOYSREQUEST']._serialized_start=1029
  _globals['_GETENVOYSREQUEST']._serialized_end=1047
  _globals['_GETENVOYSRESPONSE']._serialized_start=1049
  _globals['_GETENVOYSRESPONSE']._serialized_end=1139
  _globals['_GETFLOWSTATEREQUEST']._serialized_start=1141
  _globals['_GETFLOWSTATEREQUEST']._serialized_end=1162
  _globals['_GETFLOWSTATERESPONSE']._serialized_start=1164
  _globals['_GETFLOWSTATERESPONSE']._serialized_end=1225
  _globals['_SENDRUNTIMEREQUEST']._serialized_start=1227
  _globals['_SENDRUNTIMEREQUEST']._serialized_end=1247
  _globals['_RUNTIMEREQUESTRESPONSE']._serialized_start=1249
  _globals['_RUNTIMEREQUESTRESPONSE']._serialized_end=1291
  _globals['_GETEXPERIMENTSTDOUTREQUEST']._serialized_start=1293
  _globals['_GETEXPERIMENTSTDOUTREQUEST']._serialized_end=1346
  _globals['_GETEXPERIMENTSTDOUTRESPONSE']._serialized_start=1348
  _globals['_GETEXPERIMENTSTDOUTRESPONSE']._serialized_end=1441
  _globals['_DIRECTOR']._serialized_start=1444
  _globals['_DIRECTOR']._serialized_end=2770
# @@protoc_insertion_point(module_scope)
